
import datetime
import sys
from pathlib import Path
import json
import click
import traceback

from .utils import (
    InvalidProjectGroupError,
    OutsideProjectPathError,
    get_parents_path,
    get_project_group,
)
from damply.project import DirectoryAudit
from damply import __version__ as damply_version
from damply.logging_config import logger

@click.option(
    '--force-compute-details',
    '-f',
    'compute_details',
    help='Force the computation of details for the directory and subdirectories regardless of cache.',
    is_flag=True,
    default=False,
    show_default=True,
)
@click.argument("directory", type=click.Path(exists=True, file_okay=False, path_type=Path))
@click.command("audit")
def audit(directory: Path, compute_details: bool) -> None:
    """Audit all subdirectories and aggregate damply output into a single JSON.
    
    Unlike the 'damply proejct' command, this will by default, try to compute
    details, using cache if it exists. If you want to force the computation
    of details for the directory and subdirectories, use the --force-compute-details
    flag. This will ignore any cached results and recompute everything.

    """

    # resolve the directory to scan
    directory = directory.expanduser().resolve().absolute()

    logger.info(f"Starting audit for directory: {directory}")

    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%dT%H-%M-%S")
    date = now.strftime("%Y-%m-%d")

    try:
        project_group = get_project_group(directory)
        relative_path = get_parents_path(project_group, directory)
    except (InvalidProjectGroupError, OutsideProjectPathError) as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)

    results_dir = Path(f"/cluster/projects/{project_group}/admin/audit/results/{date}/{relative_path}")
    results_dir / "audit.json"

    logger.debug(
        f"Creating results directory: {results_dir} "
        f"for project group: {project_group}, relative path: {relative_path}"
    )
    results_dir.mkdir(parents=True, exist_ok=True)
    output_file = results_dir / "audit.json"

    # Initialize combined results
    results = {
        "audit_date": timestamp,
        "source_directory": {},
        "directories": {},
        "damply_version": damply_version,
    }
    
    def safe_audit(path: Path) -> dict:
        try:
            logger.info(f"Auditing: {path}")
            audit_obj = DirectoryAudit.from_path(path)
            audit_obj.compute_details(show_progress=True, force=compute_details)
            return {
                "status": "ok",
                "data": audit_obj.to_dict(),
            }
        except Exception as e:
            etype = type(e).__name__
            emsg = str(e)
            stack = traceback.format_exc(limit=5)  # limit frames if needed
            logger.exception(f"Failed audit: {path}")

            return {
                "status": "error",
                "error": f"{etype}: {emsg}",
                "traceback": stack,
            }

    # Audit subdirectories
    for subdir in sorted(directory.iterdir()):
        if subdir.is_dir():
            results["directories"][subdir.name] = safe_audit(subdir)

    # count number of errors
    logger.info(f"Total directories audited: {len(results['directories'])}")
    error_count = sum(
        1 for result in results["directories"].values() if result["status"] == "error"
    )
    if error_count > 0:
        logger.error(f"Total errors encountered: {error_count}")


    # Audit the main source directory
    results["source_directory"] = safe_audit(directory)

    # Save combined results
    output_file.write_text(json.dumps(results, default=str, indent=2))
    logger.info(f"Audit results saved to: {output_file}")





from __future__ import annotations

import subprocess
from pathlib import Path

from bytesize import ByteSize
from rich.progress import Progress, SpinnerColumn, TextColumn

from .find_readme import find_readme, parse_readme

__all__ = [
	'get_directory_size',
	'count_files',
	'collect_suffixes',
	'find_readme',
	'parse_readme',
]


def get_directory_size(directory: Path, show_progress: bool = True) -> ByteSize:
	if show_progress:
		with Progress(
			SpinnerColumn(),
			TextColumn('[progress.description]{task.description}'),
			transient=True,  # This makes the progress bar disappear after completion
		) as progress:
			task = progress.add_task(
				f'Calculating size of {str(directory.absolute())}...', total=None
			)
			result = subprocess.run(
				['du', '-s', '-B 1', str(directory)],
				capture_output=True,
				text=True,
				check=True,
			)
			progress.update(task, completed=True)
	else:
		result = subprocess.run(
			['du', '-s', '-B 1', str(directory)],
			capture_output=True,
			text=True,
			check=True,
		)

	size_ = ByteSize(int(result.stdout.split()[0]))
	return size_


def count_files(directory: Path, show_progress: bool = True) -> int:
	count = 0

	with Progress(
		SpinnerColumn(),
		TextColumn('[progress.description]{task.description}'),
		transient=True,
		disable=not show_progress,
	) as progress:
		task = progress.add_task(
			f'Counting files in {str(directory.absolute())}...', total=None
		)

		for path in directory.rglob('*'):
			if path.is_file():
				count += 1

		progress.update(task, completed=True)

	return count


def collect_suffixes(directory: Path, show_progress: bool = True) -> list[str]:
	suffixes: set[str] = set()

	with Progress(
		SpinnerColumn(),
		TextColumn('[progress.description]{task.description}'),
		transient=True,
		disable=not show_progress,
	) as progress:
		task = progress.add_task(
			f'Collecting file extensions in {str(directory.absolute())}...', total=None
		)

		for path in directory.rglob('*'):
			if (
				path.is_file()
				and not path.name.startswith('.')
				and path.suffixes
				and len(path.suffixes) <= 3  # 4 suffixes is crazy
			):
				valid_suffixes = [
					suffix
					for suffix in path.suffixes
					if not any(
						c.isdigit() for c in suffix
					)  # ignore suffixes with digits
				]
				suffixes.add(''.join(valid_suffixes))

		progress.update(task, completed=True)

	return list(suffixes)


if __name__ == '__main__':
	import sys

	if len(sys.argv) != 2:
		sys.exit(1)
	directory_path = Path(sys.argv[1])
	if not directory_path.is_dir():
		sys.exit(1)
	size = get_directory_size(directory_path, show_progress=True)



from pathlib import Path

import rich_click as click
from rich import print

from damply.cli.click_config import help_config
from damply.metadata import MANDATORY_FIELDS, DMPMetadata


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'path',
	type=click.Path(
		exists=False,
		path_type=Path,
		file_okay=False,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.rich_config(help_config=help_config)
def init(path: Path) -> None:
	"""Initialize a new README file."""

	if not path.exists():
		print(f'WARNING: Directory {path} does not exist. Creating it now.')
		path.mkdir(parents=True)

	try:
		metadata = DMPMetadata.from_path(path)
		if metadata.readme.exists():
			print(f'Error: README file already exists at {metadata.readme}')
			return
	except ValueError:
		pass

	from rich.console import Console

	console = Console()
	new_readme_path = path / 'README.md'
	console.print(f'Creating a new README file at {new_readme_path}')

	fields = {fld: '' for fld in MANDATORY_FIELDS}
	# create the README file
	for fld in MANDATORY_FIELDS:
		while not fields[fld]:
			print(f'[red]Field {fld} MUST be set.[/red]')
			fields[fld] = console.input(f'Enter a value for {fld}: ')

	new_metadata = DMPMetadata(
		fields=fields,
		content='',
		path=path,
		permissions='---------',
		logs=[],
		readme=new_readme_path,
	)

	try:
		new_metadata.check_fields()
	except ValueError as e:
		print(f'Error: {e}')
		return

	new_metadata.write_to_file()
	console.print(new_metadata)



"""dmpdirs: Standardized access to project directories in data science workflows.

This package provides convenient access to common directories in data science projects,
supporting environment variable configuration and automatic directory creation.

Directory paths are determined in the following order:
1. From corresponding environment variables (if set)
2. Default paths relative to the project root

Environment variables that can be used:
- DMP_PROJECT_ROOT or PIXI_PROJECT_ROOT: The root directory of the project
- CONFIG: Configuration files directory
- METADATA: Dataset descriptions directory
- LOGS: Log files directory
- RAWDATA: Raw input data directory
- PROCDATA: Processed/intermediate data directory
- RESULTS: Analysis outputs directory
- SCRIPTS: Analysis scripts directory
- NOTEBOOKS: Jupyter notebooks directory

Default directory structure (when environment variables are not set):
```console
project_root/
├── config/         # Configuration files
├── data/           # All data in one parent directory
│   ├── procdata/   # Processed/intermediate data
│   ├── rawdata/    # Raw input data
│   └── results/    # Analysis outputs
├── logs/           # Log files
├── metadata/       # Dataset descriptions
└── workflow/       # Code organization
    ├── notebooks/  # Jupyter notebooks
    └── scripts/    # Analysis scripts
```

Usage:
    from damply.dmpdirs import dirs

    # Access paths as Path objects
    data_file = dirs.RAWDATA / "dataset.csv"

    # Print absolute paths
    print(dirs.RAWDATA)  # e.g., /Users/username/projects/my_project/data/rawdata

    # Disable strict mode to auto-create directories
    dirs.set_strict_mode(False)
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import ClassVar, Dict, List

from damply import logger

from .exceptions import DirectoryNameNotFoundError, EnvironmentVariableNotSetError


def get_project_root() -> Path:
	"""Get the project root directory."""
	# Check if the environment variable is set
	if project_root := os.getenv('DMP_PROJECT_ROOT') or (
		project_root := os.getenv('PIXI_PROJECT_ROOT')
	):
		return Path(project_root).resolve()

	# If not, use the current working directory
	return Path.cwd().resolve()


class DamplyDirs:
	"""Class that provides computed properties for project directories.

	This class provides computed properties that lazily evaluate directory paths
	using environment variables when available, and raising appropriate errors
	if directories don't exist when accessed.

	Directory access is provided through attribute access:
	- dirs.RAWDATA: Raw data directory
	- dirs.PROCDATA: Processed data directory
	- dirs.RESULTS: Results directory
	- dirs.CONFIG: Configuration files directory
	- dirs.METADATA: Metadata directory
	- dirs.LOGS: Log files directory
	- dirs.SCRIPTS: Scripts directory
	- dirs.NOTEBOOKS: Notebooks directory

	By default, the class operates in "strict" mode, which means it will raise an error
	if a requested directory doesn't exist. This can be changed using the `set_strict_mode()`
	method.
	"""

	# Class variable to hold singleton instance
	_instance: ClassVar[DamplyDirs | None] = None

	# List of valid directory names
	_VALID_DIRS: ClassVar[List[str]] = [
		'PROJECT_ROOT',
		'RAWDATA',
		'PROCDATA',
		'RESULTS',
		'METADATA',
		'LOGS',
		'CONFIG',
		'SCRIPTS',
		'NOTEBOOKS',
	]

	# Private instance attributes
	_initialized: bool
	_project_root: Path
	_dir_cache: Dict[str, Path]
	_strict: bool = False  # Controls whether directories are created when missing

	def __new__(cls) -> DamplyDirs:
		"""Implement singleton pattern."""
		if cls._instance is None:
			cls._instance = super(DamplyDirs, cls).__new__(cls)
			cls._instance._initialized = False
		return cls._instance

	def __init__(self) -> None:
		"""Initialize the DamplyDirs object (singleton)."""
		# Skip initialization if already initialized
		if getattr(self, '_initialized', False):
			return

		# Initialize core attributes
		self._project_root = get_project_root()
		self._dir_cache: Dict[str, Path] = {}
		self._initialized = True

	def set_strict_mode(self, strict: bool) -> None:
		"""Set the strict mode for directory creation.

		In strict mode, errors are raised if directories don't exist.
		Otherwise, missing directories are created automatically.

		Args:
		    strict: If True, enable strict mode. If False, disable strict mode.
		"""
		self._strict = strict

	def _get_dir_path(self, dir_name: str) -> Path:  # noqa: PLR0912
		"""Get the path for a directory based on environment variables or default paths.

		Args:
		    dir_name: The name of the directory to get

		Returns:
		    Path object for the requested directory

		Raises:
		    DirectoryNameNotFoundError: If the directory doesn't exist and strict mode is enabled
		    EnvironmentVariableNotSetError: If env variable check is requested but not found
		"""
		# Return from cache if available
		if dir_name in self._dir_cache:
			return self._dir_cache[dir_name]

		# Special case for PROJECT_ROOT
		if dir_name == 'PROJECT_ROOT':
			return self._project_root

		# Check if an environment variable with the same name exists
		if env_path := os.getenv(dir_name):
			# Use the environment variable value
			path = Path(env_path).resolve()
		else:
			# if strict mode is on, raise an error that environment variable is not set
			if self._strict:
				raise EnvironmentVariableNotSetError(dir_name)
			else:
				# If not strict, use default paths
				logger.warning(
					f"Environment variable '{dir_name}' is not set. "
					'Using default path relative to project root.'
				)

			# Fall back to default paths relative to project root
			if dir_name in ['RAWDATA', 'PROCDATA', 'RESULTS']:
				# Data directories are in data/ subdirectory
				data_dir = self._project_root / 'data'
				path = data_dir / dir_name.lower()
			elif dir_name in ['SCRIPTS', 'NOTEBOOKS']:
				# Code directories are in workflow/ subdirectory
				workflow_dir = self._project_root / 'workflow'
				path = workflow_dir / dir_name.lower()
			else:
				# Other directories are directly in the project root
				path = self._project_root / dir_name.lower()

		# mkdir if not strict mode and path doesn't exist
		if not path.exists() and not self._strict:
			try:
				logger.info(f"Creating directory '{path}' because it does not exist.")
				path.mkdir(parents=True, exist_ok=True)
			except Exception as e:
				raise DirectoryNameNotFoundError(dir_name, str(path)) from e

		# Cache the path
		self._dir_cache[dir_name] = path

		# Check if directory exists and create it or raise error if needed
		if not path.exists():
			if self._strict:
				raise DirectoryNameNotFoundError(dir_name, str(path))
			else:
				# Create the directory if it doesn't exist and we're in non-strict mode
				path.mkdir(parents=True, exist_ok=True)

		return path

	def __getattr__(self, name: str) -> Path:
		"""Get attribute for a directory name.

		Args:
		    name: The name of the directory to get

		Returns:
		    Path object for the requested directory

		Raises:
		    AttributeError: If the attribute is not a recognized directory
		    DirectoryNameNotFoundError: If the directory doesn't exist and strict mode is on
		"""
		if name.isupper() and name in self._VALID_DIRS:
			return self._get_dir_path(name)
		errmsg = (
			f"'{name}' is not a valid directory name. "
			f'Valid names are: {", ".join(self._VALID_DIRS)}'
		)
		raise AttributeError(errmsg)

	def __getitem__(self, key: str) -> Path:
		"""Support dictionary-style access to directories.

		Args:
		    key: Directory name (e.g., "RAWDATA", "CONFIG")

		Returns:
		    Path object for the requested directory

		Raises:
		    KeyError: If the key is not a recognized directory
		    DirectoryNameNotFoundError: If the directory doesn't exist and strict mode is on
		"""
		try:
			return getattr(self, key)
		except AttributeError as e:
			raise KeyError(str(e)) from e

	def __dir__(self) -> List[str]:
		"""Return list of available attributes for tab completion."""
		return self._VALID_DIRS + ['set_strict_mode']

	def __repr__(self) -> str:
		"""Return a tree-like representation of the directory structure."""
		strict_mode = f'Strict Mode: {"ON" if self._strict else "OFF"}'
		root_info = f'Project Root: {self._project_root}'

		# Create tree structure
		tree = [f'DamplyDirs<{strict_mode}>', root_info]

		for dir_name in sorted(
			dir_name for dir_name in self.__dir__() if dir_name.isupper()
		):
			if dir_name != 'PROJECT_ROOT':
				try:
					path = getattr(self, dir_name)
					rel_path = path.relative_to(self._project_root)
					tree.append(f'{dir_name:<13}: ├── {rel_path}')
				except DirectoryNameNotFoundError:
					tree.append(f'{dir_name:<13}: ├── <not found>')
				except ValueError:
					# For paths that can't be made relative (e.g., on different drives)
					tree.append(f'{dir_name:<13}: ├── {path} (absolute)')

		# Fix the last item to use └── instead of ├──
		if len(tree) > 2:
			tree[-1] = tree[-1].replace('├──', '└──')

		return '\n'.join(tree)


# Create a singleton instance that will be imported by users
dirs = DamplyDirs()



import pathlib

import click


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option(
	'--force-compute-details',
	'-f',
	'compute_details',
	help='Display extra information about the project, i.e size. Takes longer to compute.',
	is_flag=True,
	default=False,
	show_default=True,
)
@click.argument(
	'directory',
	type=click.Path(
		exists=True,
		file_okay=False,
		dir_okay=True,
		resolve_path=True,
		path_type=pathlib.Path,
	),
	default=None,
	required=False,
)
@click.option(
	'--json',
	'-j',
	is_flag=True,
	help='Output the audit information in JSON format.',
)
def project(
	directory: pathlib.Path | None,
	json: bool,
	compute_details: bool,
	force: bool = False,
) -> None:
	"""Display information about the current project.

	DIRECTORY is the path to the project directory,
	if not provided, the current working directory will be used.

	This command caches directory size and file count calculations to improve performance.
	By default, it will use cached values if available and if the directory hasn't been
	modified since the cache was created. Use --force to bypass the cache and recalculate
	all values.
	"""
	from damply.project import DirectoryAudit

	if directory is None:
		directory = pathlib.Path.cwd()
	directory = directory.resolve()
	audit = DirectoryAudit.from_path(directory)

	if compute_details:
		audit.compute_details(show_progress=True, force=True)

	if json:
		print(audit.to_json())  # noqa
	else:
		from rich import print as rprint

		rprint(audit)



import stat
from dataclasses import asdict, dataclass, field
from datetime import datetime
from pathlib import Path
from zoneinfo import ZoneInfo

import rich.repr
from bytesize import ByteSize
from rich import print

from damply import __version__
from damply.cache_dir import get_cache_dir
from damply.logging_config import logger
from damply.utils import (
	collect_suffixes,
	count_files,
	find_readme,
	get_directory_size,
	parse_readme,
)


@dataclass
class DirectoryAudit:
	path: Path
	owner: str
	group: str
	full_name: str
	permissions: str
	last_modified: datetime
	last_changed: datetime
	readme_path: Path | None = field(default=None)
	metadata: dict[str, str] = field(default_factory=dict, init=False)

	# lazy evaluated because expensive
	size: ByteSize | None = field(default=None, init=False)
	file_count: int | None = field(default=None, init=False)
	file_types: list[str] | None = field(default=None, init=False)

	last_computed: str | None = field(default=None, init=False, repr=False)

	def __post_init__(self) -> None:
		# parse readme file if it exists
		if not self.readme_path:
			return
		self.metadata = parse_readme(self.readme_path)

	@classmethod
	def from_path(cls, path: Path) -> 'DirectoryAudit':
		# resolve ~ and expand environment variables and canonicalize the path
		path = path.expanduser().resolve()
		if not path.exists():
			msg = f'The path {path} does not exist.'
			raise FileNotFoundError(msg)
		elif not path.is_dir():
			msg = f'The path {path} is not a directory.'
			raise NotADirectoryError(msg)

		stats = path.stat()

		try:
			from grp import getgrgid
			from pwd import getpwuid

			pwuid_name = getpwuid(stats.st_uid).pw_name
			pwuid_gecos = getpwuid(stats.st_uid).pw_gecos
			group_name = getgrgid(stats.st_gid).gr_name
		except ImportError:
			pwuid_name = 'Unknown'
			pwuid_gecos = 'Unknown'
			group_name = 'Unknown'
		except KeyError:
			pwuid_name = 'Unknown'
			pwuid_gecos = 'Unknown'
			group_name = 'Unknown'

		toronto_tz = ZoneInfo('America/Toronto')
		audit = DirectoryAudit(
			path=path,
			owner=pwuid_name,
			group=group_name,
			full_name=pwuid_gecos,
			permissions=stat.filemode(stats.st_mode),
			last_modified=datetime.fromtimestamp(stats.st_mtime, tz=toronto_tz),
			last_changed=datetime.fromtimestamp(stats.st_ctime, tz=toronto_tz),
			readme_path=find_readme(path),
		)
		return audit

	def compute_details(self, show_progress: bool = True, force: bool = False) -> None:
		"""Compute size and file count, using cache if available and up-to-date.

		Args:
		    show_progress: Whether to show progress bars during computation
		    force: If True, recompute regardless of cache
		"""
		# Check if we have cached data and whether it's still valid
		if force:
			logger.debug('Force recompute requested, ignoring cache.')
			# delete the cache if it exists
			cache_path = self._get_cache_path()
			if cache_path.exists():
				logger.debug(f'Deleting cache file: {cache_path}')
				cache_path.unlink()

			# remove cached values
			self.size = None
			self.file_count = None
			self.file_types = None
		elif self._get_from_cache():
			logger.debug('Cache loaded successfully.')
			return
		else:
			logger.debug('Cache miss.')

		# Compute the values and cache them
		self.compute_size(show_progress=show_progress)
		self.compute_file_count(show_progress=show_progress)
		self.compute_file_types(show_progress=show_progress)

		self.last_computed = datetime.now(ZoneInfo('America/Toronto')).strftime(
			'%Y-%m-%d %H:%M:%S'
		)
		self._save_to_cache()

	def compute_file_count(self, show_progress: bool = True) -> int:
		"""Count files in the directory and return the count."""
		if self.file_count is None:
			self.file_count = count_files(
				directory=self.path, show_progress=show_progress
			)
		return self.file_count

	def compute_file_types(self, show_progress: bool = True) -> set[str]:
		"""Get unique file types in the directory."""
		if self.file_types is None:
			self.file_types = collect_suffixes(
				directory=self.path, show_progress=show_progress
			)
		return self.file_types

	def compute_size(self, show_progress: bool = True) -> ByteSize:
		if self.size is None:
			self.size = get_directory_size(
				directory=self.path, show_progress=show_progress
			)
		return self.size

	def _get_cache_path(self) -> Path:
		"""Get the path for the cache file for this directory."""
		cache_dir = get_cache_dir() / 'directory_audits'
		cache_dir.mkdir(parents=True, exist_ok=True)

		# Create a unique filename based on the path
		path_str = str(self.path.absolute())
		import hashlib

		path_hash = hashlib.md5(path_str.encode()).hexdigest()
		return cache_dir / f'{path_hash}.json'

	def _save_to_cache(self) -> None:
		"""Save the current state to cache."""
		cache_path = self._get_cache_path()
		import json

		# Create a dict with all the necessary data
		cache_data = {
			'path': str(self.path.absolute()),
			'last_computed': self.last_computed,
			'size': str(self.size) if self.size is not None else None,
			'file_count': self.file_count,
			'file_types': self.file_types if self.file_types else None,
			'last_modified': self.last_modified.isoformat()
			if self.last_modified
			else None,
			# 'last_accessed': self.last_accessed.isoformat()
			# if self.last_accessed
			# else None,
			'last_changed': self.last_changed.isoformat()
			if self.last_changed
			else None,
			'damply_version': __version__,  # Save the current version
		}

		with cache_path.open('w', encoding='utf-8') as f:
			json.dump(cache_data, f)

	def _get_from_cache(self) -> bool:
		"""
		Try to load data from cache if it's newer than the last modified date.

		Returns:
		    bool: True if cache was loaded successfully, False otherwise
		"""
		cache_path = self._get_cache_path()
		if not cache_path.exists():
			return False

		import json

		try:
			with cache_path.open('r', encoding='utf-8') as f:
				cache_data = json.load(f)

			from packaging import version

			# Check version compatibility
			cached_version = cache_data.get('damply_version', None)
			if cached_version is None or version.parse(cached_version) < version.parse(
				__version__
			):
				logger.debug(
					f'Cache version {cached_version} is older than current version {__version__}.'
				)
				# Delete the outdated cache
				cache_path.unlink()
				return False

			# Check if the cache is for the same path
			# this would be wild to not be the same right now
			if cache_data.get('path') != str(self.path.absolute()):
				return False

			# Check if the directory has been modified since the cache was created
			cached_last_modified = datetime.fromisoformat(
				cache_data.get('last_modified')
			)

			# Ensure we're comparing datetimes with the same timezone awareness
			if (
				cached_last_modified.tzinfo is None
				and self.last_modified.tzinfo is not None
			):
				# Convert naive to aware using the same timezone as self.last_modified
				cached_last_modified = cached_last_modified.replace(
					tzinfo=self.last_modified.tzinfo
				)
			elif (
				cached_last_modified.tzinfo is not None
				and self.last_modified.tzinfo is None
			):
				# Convert self.last_modified to aware using cached timezone
				self.last_modified = self.last_modified.replace(
					tzinfo=cached_last_modified.tzinfo
				)

			if self.last_modified > cached_last_modified:
				return False

			# Load the cached values
			if cache_data.get('size') is not None:
				self.size = ByteSize(cache_data['size'])
			self.file_count = cache_data.get('file_count')
			self.last_computed = cache_data.get('last_computed')
			self.file_types = cache_data.get('file_types', set())
		except (json.JSONDecodeError, KeyError, ValueError) as e:
			logger.debug(f'Error loading cache: {e}')
			# If anything goes wrong, delete cache and recompute
			if cache_path.exists():
				cache_path.unlink()
			return False
		else:
			return True

	def __rich_repr__(self) -> rich.repr.Result:
		yield 'path', self.path.absolute()
		yield 'owner', self.owner
		yield 'group', self.group
		yield 'full_name', self.full_name
		yield 'permissions', self.permissions
		# yield 'last_accessed', self.last_accessed.date() if self.last_accessed else None
		yield 'last_modified', self.last_modified.date() if self.last_modified else None
		yield 'last_changed', self.last_changed.date() if self.last_changed else None
		yield 'size', self.size if self.size is not None else 'Not computed yet'
		yield (
			'file_count',
			self.file_count if self.file_count is not None else 'Not computed yet',
		)
		yield (
			'file_types',
			self.file_types if self.file_types is not None else 'Not computed yet',
		)
		yield 'readme_path', self.readme_path if self.readme_path else 'No README found'
		yield (
			'metadata',
			self.metadata if self.metadata else 'No metadata found in README',
		)

	def to_dict(self) -> dict:
		result = {}
		for k, v in asdict(self).items():
			if k.startswith('_'):
				continue

			# Convert datetime to date for display purposes
			if isinstance(v, datetime):
				result[k] = v.date()
			else:
				result[k] = v
		return result

	def to_json(self, indent: int = 4) -> str:
		import json

		return json.dumps(self.to_dict(), default=str, indent=indent)


if __name__ == '__main__':
	from rich import print

	audit = DirectoryAudit.from_path(
		Path('/cluster/projects/radiomics/Projects/testimgtools_benchmark/results')
	)
	print(audit)
	audit.compute_details(show_progress=True, force=True)

	print('*' * 20)
	print(audit.to_dict())
	print('*' * 20)
	print(audit.to_json())

	print('*' * 20)
	print(f'Size: {audit.size:.2f:GB}')

	print('*' * 20)
	print(audit.to_dict())



import os
import sys

from loguru import logger

# Read environment variable or default to WARNING
log_level = os.getenv('DAMPLY_LOG_LEVEL', 'WARNING').upper()

# change the log level of the logger
logger.remove()  # Remove the default logger
logger.add(
	sys.stderr,
	level=log_level,
	format='<green>{time:YYYY-MM-DD}</green> | <level>{level}</level> | {message}',
	colorize=True,
)



"""helper function to find and parse the README file in a directory.

The README can either be in the root OR in a subdirectory called 'docs'.

also a fun
"""

import re
from pathlib import Path

from damply.logging_config import logger


def find_readme(directory: Path) -> Path | None:
	"""Find any README file in the given directory or its 'docs' subdirectory."""

	readmes = [
		f for f in directory.glob('readme*', case_sensitive=False) if f.is_file()
	] + [
		f
		for f in (directory / 'docs').glob('readme*', case_sensitive=False)
		if f.is_file()
	]

	if not readmes:
		return None

	if len(readmes) == 1:
		return readmes[0]

	# if there are multiple readmes, just prefer the the first one
	logger.warning(
		f'Found {len(readmes)} README files in {directory}. '
		'Returning the first one found. '
	)
	return readmes[0]


def parse_readme(
	readme: Path,
	pattern: re.Pattern[str] = re.compile(r'^#([A-Z]+): (.+)$'),
) -> dict[str, str]:
	"""Parse the README file and return its content as a string."""
	current_field = None
	current_value = []
	content_lines = []  # everything else that is not a field
	metadata = {}
	with readme.open(mode='r') as file:
		for line in file:
			if line.strip() == '' and current_field:
				# End current field on double newline
				metadata[current_field] = ' '.join(current_value).strip()
				current_field = None
				current_value = []
			else:
				match = pattern.match(line.strip())
				if match:
					if current_field:
						metadata[current_field] = ' '.join(current_value).strip()
					current_field, current_value = (
						match.groups()[0],
						[match.groups()[1]],
					)
				elif current_field:
					current_value.append(line.strip())
				else:
					content_lines.append(line.strip())

		if current_field:
			metadata[current_field] = ' '.join(current_value).strip()

	# were gonna ignore the content lines for now
	# metadata['content'] = '\n'.join(content_lines).strip()
	return metadata


if __name__ == '__main__':
	import click

	@click.command()
	@click.argument(
		'directory', type=click.Path(exists=True, file_okay=False, dir_okay=True)
	)
	def main(directory: str) -> None:
		"""Find and parse the README file in the given directory."""
		readme_path = find_readme(Path(directory))
		if readme_path:
			metadata = parse_readme(readme_path)
			for _key, _value in metadata.items():
				pass
		else:
			pass

	main()



"""Custom exceptions for the dmpdirs package."""


class DirectoryNameNotFoundError(Exception):
	"""Exception raised when a required project directory name is not found in the config."""

	def __init__(self, directory_name: str, directory_path: str) -> None:
		self.directory_name = directory_name
		message = (
			f"Project directory name '{directory_name}' not found in configuration"
			f" or does not exist at '{directory_path}'"
		)
		super().__init__(message)


class EnvironmentVariableNotSetError(Exception):
	"""Exception raised when a required environment variable is not set."""

	def __init__(self, var_name: str) -> None:
		self.var_name = var_name
		message = (
			f"Environment variable '{var_name}' is not set. "
			f'Please set it or use non-strict mode to use default paths.'
		)
		super().__init__(message)



from pathlib import Path
from typing import Union


class InvalidProjectGroupError(ValueError):
	"""Raised when the project group is not 'bhklab' or 'radiomics'."""


class OutsideProjectPathError(ValueError):
	"""Raised when the source directory is not within the expected project base path."""


def get_project_group(source_dir: Union[str, Path]) -> str:
	"""Extract the project group from the source directory path.

	Args:
	    source_dir: Path to the source directory.

	Returns:
	    The project group name ('bhklab' or 'radiomics').

	Raises:
	    InvalidProjectGroupError: If the group is not bhklab or radiomics.
	    OutsideProjectPathError: If the path doesn't start with /cluster/projects.

	Examples:
	    >>> get_project_group('/cluster/projects/bhklab/rawdata/CCLE')
	    'bhklab'

	    >>> get_project_group(Path('/cluster/projects/radiomics/PublicDataSets'))
	    'radiomics'
	"""
	source_path = Path(source_dir).resolve()
	parts = source_path.parts

	if len(parts) < 4 or parts[1] != 'cluster' or parts[2] != 'projects':
		msg = (
			'Input directory must start with /cluster/projects/bhklab '
			'or /cluster/projects/radiomics.'
		)
		raise OutsideProjectPathError(msg)

	project_group = parts[3]
	if project_group not in {'bhklab', 'radiomics'}:
		msg = (
			f"Invalid project group '{project_group}'. Must be 'bhklab' or 'radiomics'."
		)
		raise InvalidProjectGroupError(msg)

	return project_group


def get_parents_path(project_group: str, source_dir: Union[str, Path]) -> str:
	"""Compute the relative path from the project root to the source directory.

	Args:
	    project_group: 'bhklab' or 'radiomics'.
	    source_dir: Full path to the source directory.

	Returns:
	    A string path relative to the group root.

	Raises:
	    OutsideProjectPathError: If the path is not under the group base path.

	Examples:
	    >>> get_parents_path('bhklab', '/cluster/projects/bhklab/rawdata/CCLE')
	    'rawdata/CCLE'
	"""
	base_path = Path(f'/cluster/projects/{project_group}').resolve()
	source_path = Path(source_dir).resolve()

	try:
		return str(source_path.relative_to(base_path))
	except ValueError as e:
		msg = f'{source_path} is not within {base_path}'
		raise OutsideProjectPathError(msg) from e



import rich_click as click

click.rich_click.STYLE_OPTIONS_TABLE_BOX = 'SIMPLE'
click.rich_click.STYLE_COMMANDS_TABLE_SHOW_LINES = True
click.rich_click.STYLE_COMMANDS_TABLE_PAD_EDGE = True


click.rich_click.OPTION_GROUPS = {
	'damply': [
		{
			'name': 'Basic options',
			'options': ['--help', '--version'],
		},
	]
}

click.rich_click.COMMAND_GROUPS = {
	'damply': [
		{
			'name': 'Info Commands',
			'commands': ['view', 'whose', 'add-field', 'log', 'config', 'init', 'size'],
		},
		{
			'name': 'Audit Commands',
			'commands': [
				'audit',
				'plot',
			],
		},
	]
}


help_config = click.RichHelpConfiguration(
	show_arguments=True,
	option_groups={'damply': [{'name': 'Arguments', 'panel_styles': {'box': 'ASCII'}}]},
)



"""CLI tool to take a list of group names on the system, get their users, and print a table.

i.e given groups 'bhklab, radiomics, bhklab_icb' it will print a table of users in those groups.
columns would be:

- username
- realname
- userid
- bhklab
- radiomics
- bhklab_icb

and then in each row, it will have a 1 if the user is in that group, and a 0 if not.

"""

from typing import TYPE_CHECKING, Sequence

import click

from damply import logger

if TYPE_CHECKING:
	import pandas as pd

DEFAULT_GROUPS = [
	'bhklab',
	'radiomics',
	'ega',
	'bhklab_icb',
	'bhklab_pmcc_gyn_autosegmentation',
	'bhklab_pmcc_gyn_gu',
	'pmdatascience',
]


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option(
	'--additional-groups',
	'-a',
	'additional_groups',
	help='Additional comma separated groups to the default list.',
	default=None,
)
@click.option(
	'--keep-extra-groups',
	'-k',
	'keep_extra_groups',
	is_flag=True,
	default=False,
	help='Keep columns for groups not explicitly requested.',
)
@click.option(
	'--csv',
	is_flag=True,
	default=False,
	help='Output the table as a CSV file instead of printing it, useful for further processing or > into files.',
)
@click.argument('group_names', nargs=-1, required=False)
def groups_table(
	group_names: list[str] | None = None,
	additional_groups: str | None = None,
	keep_extra_groups: bool = False,
	csv: bool = False,
) -> None:
	"""
	Generate a user-group membership table from group names.

	This CLI tool takes one or more Unix/Linux group names, collects their member users,
	enriches that information with full user metadata (UID, real name, etc.), and outputs
	a formatted table where each row represents a user and each group has its own column.

	By default, the tool uses a set of lab-specific groups but can be customized to include
	others. The resulting table indicates with a 1 or 0 whether a user is a member of each group.

	Default groups are: bhklab, radiomics, bhklab_icb, bhklab_pmcc_gyn_autosegmentation, bhklab_pmcc_gyn_gu.

	\b
	Examples
	--------
	Basic usage with default group set:

	    $ groups-table

	Include extra groups beyond the default:

	    $ groups-table --additional-groups ega,cbmp

	Fully custom group list (ignores defaults):

	    $ groups-table bhklab radiomics cbmp

	Keep other group columns found in users' metadata:

	    $ groups-table bhklab radiomics --keep-extra-groups

	Export as CSV:

	    $ groups-table bhklab radiomics --csv > output.csv

	\b
	Notes
	-----
	This tool requires access to system group/user information, and may rely on NSS/SSSD/LDAP to
	resolve group memberships. If some groups are not enumerable, users in those groups may be
	resolved indirectly by collecting members from requested groups.
	"""
	logger.debug('Starting groups_table CLI tool...')
	from itertools import chain

	from rich import print

	from damply.admin import GroupInfo, UserInfo

	if not group_names:
		group_names = DEFAULT_GROUPS

	if additional_groups:
		group_names.extend(additional_groups.split(','))

	logger.debug('Fetching group information...')
	group_info_db = [GroupInfo.from_group_name(group) for group in group_names]

	all_usernames = sorted(set(chain.from_iterable(g.members for g in group_info_db)))

	user_info_db: list[UserInfo]
	logger.debug('Fetching user information...')
	user_info_db = [UserInfo.from_username(username) for username in all_usernames]

	# print([u.to_dict() for u in user_info_db])
	user_dicts = [u.to_dict() for u in user_info_db]

	group_data_df = build_group_membership_table(
		user_dicts=user_dicts,
		requested_groups=group_names,
		keep_extra_groups=keep_extra_groups,
	)

	if csv:
		logger.debug('Outputting table as CSV...')
		# print the csv to stdout
		print(group_data_df.to_csv(index=False))
	else:
		# Pretty-print the DataFrame using Rich
		from rich.console import Console
		from rich.table import Table

		console = Console()
		table = Table(show_lines=True)
		for col in group_data_df.columns:
			table.add_column(str(col))

		for _, row in group_data_df.iterrows():
			table.add_row(*[str(x) for x in row])

		console.print(table)


def build_group_membership_table(
	user_dicts: list[dict],
	requested_groups: Sequence[str],
	keep_extra_groups: bool = False,
) -> 'pd.DataFrame':
	"""
	Processes a list of user dictionaries to create a table showing which
	users belong to which groups.
	Each group becomes a column with binary (0/1) values indicating membership.

	Parameters
	----------
	user_dicts : list[dict]
	    List of user dictionaries containing user metadata and group memberships.
	    Each dict must have 'name', 'realname', 'uid', 'default_group' and optional 'groups' keys.
	requested_groups : Sequence[str]
	    Sequence of group names to include in the output table.
	keep_extra_groups : bool, default=False
	    If True, includes all groups found in user_dicts in addition to requested_groups.
	    If False, only includes requested_groups.
	Returns
	-------
	pd.DataFrame
	    DataFrame with user metadata columns (name, realname, uid, default_group) and
	    binary group membership columns (1 for membership, 0 for non-membership).
	"""
	import pandas as pd

	all_requested_groups = set(requested_groups)

	logger.debug('Building group membership table...')

	# Start with the base user metadata
	base_df = pd.DataFrame(user_dicts)
	base_df = base_df[['name', 'realname', 'uid', 'default_group']]

	# Build group presence DataFrame
	group_flags_df = pd.DataFrame.from_records(
		[
			{**{g: 1 for g in user.get('groups', [])}, 'name': user['name']}
			for user in user_dicts
		]
	).fillna(0)

	# Merge without duplicating base metadata
	group_data_df = base_df.merge(group_flags_df, on='name', how='left').fillna(0)

	# Identify group columns (exclude base ones)
	group_cols = [
		c
		for c in group_data_df.columns
		if c not in {'name', 'realname', 'uid', 'default_group'}
	]

	# Ensure binary (int) for group flags
	group_data_df[group_cols] = group_data_df[group_cols].astype(int)

	# Control which group columns to keep
	if not keep_extra_groups:
		cols = ['name', 'realname', 'uid', 'default_group'] + sorted(
			all_requested_groups
		)
	else:
		requested = [g for g in requested_groups if g in group_cols]
		others = sorted(set(group_cols) - set(requested))
		cols = ['name', 'realname', 'uid', 'default_group'] + requested + others

	group_data_df = group_data_df[cols]

	return group_data_df



import click

from damply import __version__
from damply.cli.audit import audit
from damply.cli.groups_table import groups_table
from damply.cli.project import project
from damply.cli.whose import whose


@click.group(
	context_settings={'help_option_names': ['-h', '--help']}, no_args_is_help=True
)
@click.version_option(__version__, prog_name='damply')
def cli() -> None:
	"""
	A tool to interact with systems implementing the
	Data Management Plan (DMP) standard.

	This tool is meant to allow sys-admins to easily query and audit the metadata of their
	projects.

	To enable logging, set the env variable `DAMPLY_LOG_LEVEL`.
	"""
	pass


cli.add_command(groups_table)
cli.add_command(whose)
cli.add_command(project)
cli.add_command(audit)



#!/usr/bin/env python3
"""
Script to run damply project audit on all directories in a given path
and combine the results into a single JSON file
"""

import os
import sys
import json
import shutil
import subprocess
from datetime import datetime
from pathlib import Path
import click


def check_requirements():
    """Check if required tools are available."""
    # Check for damply
    try:
        result = subprocess.run(
            ["damply", "--version"], 
            capture_output=True, 
            text=True, 
            check=True
        )
        click.echo(f"{result.stdout.strip()} is available.")
    except (subprocess.SubprocessError, FileNotFoundError):
        click.echo("damply is not available or --help failed.")
        click.echo("please install with `uv tool install damply@latest --compile-bytecode` (on the login node)")
        sys.exit(1)

    # Check for jq
    try:
        result = subprocess.run(
            ["jq", "--version"], 
            capture_output=True, 
            text=True, 
            check=True
        )
        click.echo(f"jq {result.stdout.strip()} is available.")
    except (subprocess.SubprocessError, FileNotFoundError):
        click.echo("Error: jq is required for JSON processing but not found.")
        click.echo("Please install jq using your system package manager (i.e `pixi global install jq` or `apt-get install jq`)")
        sys.exit(1)


def get_project_group(source_dir):
    """Extract the project group from the source directory path."""
    parts = Path(source_dir).parts
    
    if len(parts) < 4 or parts[1] != "cluster" or parts[2] != "projects":
        click.echo("Error: Input directory must start with /cluster/projects/bhklab or /cluster/projects/radiomics.")
        sys.exit(1)
        
    project_group = parts[3]
    if project_group not in ["bhklab", "radiomics"]:
        click.echo(f"Error: Invalid project group '{project_group}'.")
        click.echo("Input directory must start with /cluster/projects/bhklab or /cluster/projects/radiomics.")
        sys.exit(1)
        
    click.echo(f"INFO: Project group is '{project_group}'")
    return project_group


def get_parents_path(project_group, source_dir):
    """Get the relative path from the project root."""
    base_path = Path(f"/cluster/projects/{project_group}")
    source_path = Path(source_dir)
    
    try:
        parents_path = source_path.relative_to(base_path)
        click.echo(f"INFO: Parents path is '{parents_path}'")
        return str(parents_path)
    except ValueError:
        click.echo("Error: Source directory is not within the project path.")
        sys.exit(1)


def run_damply_audit(directory, output_file, error_file):
    """Run damply project audit on a directory."""
    try:
        result = subprocess.run(
            ["damply", "project", "--force-compute-details", "--json", directory],
            capture_output=True,
            text=True,
            check=True
        )
        with open(output_file, 'w') as f:
            f.write(result.stdout)
        return True
    except subprocess.SubprocessError as e:
        with open(error_file, 'w') as f:
            f.write(str(e))
            if e.stderr:
                f.write("\n" + e.stderr)
        return False


def combine_json_files(output_file, source_file, json_files, audit_date):
    """Combine all JSON files into a single output file."""
    # Read source directory data
    with open(source_file, 'r') as f:
        source_data = json.load(f)
    
    # Create initial combined data
    combined_data = {
        "audit_date": audit_date,
        "source_directory": source_data,
        "directories": {}
    }
    
    # Add data from each directory
    for json_file in json_files:
        if not os.path.exists(json_file) or os.path.getsize(json_file) == 0:
            continue
            
        try:
            with open(json_file, 'r') as f:
                dir_data = json.load(f)
            
            dir_name = os.path.basename(json_file).replace('.json', '')
            combined_data["directories"][dir_name] = dir_data
        except json.JSONDecodeError:
            click.echo(f"Warning: Invalid JSON in {json_file}, skipping")
    
    # Write combined data to output file
    with open(output_file, 'w') as f:
        json.dump(combined_data, f, indent=2)


@click.command()
@click.argument('directory_to_scan', type=click.Path(exists=True, file_okay=False, dir_okay=True))
def audit_directory(directory_to_scan):
    """Run damply project audit on all directories in a given path and combine the results."""
    source_dir = os.path.realpath(directory_to_scan)
    
    # Check requirements
    check_requirements()
    
    # Get timestamps
    now = datetime.now()
    today = now.strftime("%Y-%m-%dT%H-%M-%S")
    today_date = now.strftime("%Y-%m-%d")
    
    # Get project info
    project_group = get_project_group(source_dir)
    parents_path = get_parents_path(project_group, source_dir)
    
    # Setup directories
    results_dir = Path(f"/cluster/projects/{project_group}/admin/audit/results/{today_date}/{parents_path}")
    temp_dir = results_dir / ".tmp" / f"damply_audit_{os.path.basename(source_dir)}"
    output_file = results_dir / "audit.json"
    
    click.echo(f"INFO: Results directory is '{results_dir}'")
    click.echo(f"INFO: Temporary directory is '{temp_dir}'")
    click.echo(f"INFO: Output file is '{output_file}'")
    
    # Create directories
    results_dir.mkdir(parents=True, exist_ok=True)
    temp_dir.mkdir(parents=True, exist_ok=True)
    
    # Process source directory
    click.echo("Processing source directory itself...")
    source_output = temp_dir / "source_directory.json"
    source_error = temp_dir / "source_directory.error"
    success = run_damply_audit(source_dir, source_output, source_error)
    
    if not success:
        click.echo("Error processing source directory")
        with open(source_error, 'r') as f:
            click.echo(f.read())
    else:
        if os.path.exists(source_error):
            os.unlink(source_error)
    
    # Process each subdirectory
    for item in os.listdir(source_dir):
        dir_path = os.path.join(source_dir, item)
        if os.path.isdir(dir_path):
            click.echo(f"Processing {item}...")
            dir_output = temp_dir / f"{item}.json"
            dir_error = temp_dir / f"{item}.error"
            success = run_damply_audit(dir_path, dir_output, dir_error)
            
            if not success:
                click.echo(f"Error processing {item}")
                with open(dir_error, 'r') as f:
                    click.echo(f.read())
            else:
                if os.path.exists(dir_error):
                    os.unlink(dir_error)
    
    # Find all JSON files (excluding source directory)
    json_files = [
        str(f) for f in temp_dir.glob("*.json") 
        if f.name != "source_directory.json"
    ]
    
    click.echo(f"Combining {len(json_files)} JSON files into a single object...")
    
    # Combine JSON files
    combine_json_files(
        str(output_file), 
        str(source_output), 
        json_files, 
        today
    )
    
    # Clean up temporary files
    if output_file.exists():
        click.echo("Cleaning up temporary files...")
        shutil.rmtree(temp_dir)
        
        # Remove empty directories
        for dirpath, dirnames, filenames in os.walk(results_dir, topdown=False):
            if not dirnames and not filenames and dirpath != str(results_dir):
                os.rmdir(dirpath)
    
    click.echo(f"Audit complete. Results stored in {output_file}")


if __name__ == "__main__":
    audit_directory()



from pathlib import Path

import rich_click as click
from rich import print

from damply.cli.click_config import help_config
from damply.plot import damplyplot


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
)
@click.option('--threshold_gb', type=int, default=100)
@click.option('--fig_width', type=int, default=3340)
@click.option('--fig_height', type=int, default=1440)
@click.rich_config(help_config=help_config)
def plot(
	path: Path,
	threshold_gb: int = 100,
	fig_width: int = 3340,
	fig_height: int = 1440,
) -> None:
	"""Plot the results of a damply audit using the path to the output csv file."""
	output_path = damplyplot(
		file_path=path,
		threshold_gb=threshold_gb,
		fig_width=fig_width,
		fig_height=fig_height,
	)
	print(f'The plot is saved to {output_path}')



#!/cluster/home/t119797uhn/damply/.pixi/envs/py310/bin/python 


from pathlib import Path; 
from damply.metadata import DMPMetadata
from rich import print

metadata = DMPMetadata.from_path(Path("./tests/examples/gcsi"))

print(metadata)


from pathlib import Path

from platformdirs import user_cache_dir

from damply import __author__ as package_author
from damply import __name__ as package_name


def get_cache_dir() -> Path:
	"""
	Get the path to the cache directory for the package.

	Returns:
	    Path: The path to the cache directory.
	"""
	return Path(user_cache_dir(package_name, package_author))



"""utility function to get the user info of the current user or via stat uid"""

from __future__ import annotations

import pwd
from dataclasses import asdict, dataclass


@dataclass
class UserInfo:
	"""Data class to hold user information."""

	uid: int
	name: str
	realname: str | None = None
	default_group: str | None = None

	@classmethod
	def from_username(cls, name: str) -> UserInfo:
		"""Create a UserInfo instance from a username."""
		try:
			user = pwd.getpwnam(name)
		except KeyError as ke:
			msg = f"User '{name}' does not exist."
			raise ValueError(msg) from ke

		try:
			from damply.admin.group_info import GroupInfo

			default_group = GroupInfo.from_group_id(user.pw_gid).name
		except ValueError:
			default_group = None

		return cls(
			uid=user.pw_uid,
			name=user.pw_name,
			realname=user.pw_gecos.split(',')[0] if user.pw_gecos else None,
			default_group=default_group,
		)

	@classmethod
	def from_uid(cls, uid: int) -> UserInfo:
		"""Create a UserInfo instance from a user ID."""
		try:
			user = pwd.getpwuid(uid)
		except KeyError as ke:
			msg = f'User with UID {uid} does not exist.'
			raise ValueError(msg) from ke

		try:
			from damply.admin.group_info import GroupInfo

			default_group = GroupInfo.from_group_id(user.pw_gid).name
		except ValueError:
			default_group = None

		return cls(
			uid=user.pw_uid,
			name=user.pw_name,
			realname=user.pw_gecos.split(',')[0] if user.pw_gecos else None,
			default_group=default_group,
		)

	def to_dict(self) -> dict[str, any]:
		"""Convert the UserInfo instance to a dictionary."""
		base_dict = asdict(self)
		# add groups to the dictionary
		base_dict['groups'] = self.groups
		return base_dict

	@property
	def groups(self) -> list[str]:
		"""Get the list of groups the user belongs to.

		Because of our permissions, we have to use subprocess and `groups <user>`
		to get the groups of the user.
		Returns:
		    list[str]: List of group names the user belongs to.
		"""
		import subprocess

		try:
			result = subprocess.run(
				['groups', self.name], capture_output=True, text=True, check=True
			)
			groups = result.stdout.strip().split(': ')[1].split()
		except subprocess.CalledProcessError as e:
			msg = f'Failed to get groups for user {self.name}: {e}'
			raise RuntimeError(msg) from e
		else:
			return groups

	def __repr__(self) -> str:
		return (
			f'UserInfo(uid={self.uid}, '
			f"name='{self.name}', "
			f"realname='{self.realname}', "
			f"default_group='{self.default_group}', "
			f'groups={self.groups})'
		)


if __name__ == '__main__':
	pass



import datetime  # Add this import
import re
import stat
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import Type

import rich.repr
from bytesize import ByteSize

MANDATORY_FIELDS = ['OWNER', 'DATE', 'DESC']


def is_file_writable(file_path: Path) -> bool:
	import os

	return file_path.exists() and os.access(file_path, os.W_OK)


def get_directory_size(directory: Path) -> ByteSize:
	# may raise   subprocess.CalledProcessError
	result = subprocess.run(
		['du', '-s', '-B 1', str(directory)], capture_output=True, text=True, check=True
	)
	size_ = ByteSize(int(result.stdout.split()[0]))
	return size_


@dataclass
class DMPMetadata:
	fields: dict = field(default_factory=dict)
	content: str = field(default_factory=str, repr=False)
	path: Path = field(default=Path().cwd())
	permissions: str = field(default='---------')
	logs: list = field(default_factory=list, repr=True)
	readme: Path = field(default=Path().cwd() / 'README')
	size: ByteSize = field(default=None, repr=False)
	size_measured_at: datetime.datetime = field(default=None, repr=False)

	@classmethod
	def from_path(cls: Type['DMPMetadata'], path: Path) -> 'DMPMetadata':
		readme: Path = cls._find_readme(path)

		metadata = cls()
		metadata.path = readme.resolve().parent
		metadata.permissions = cls.evaluate_permissions(readme)
		metadata.readme = readme
		if not metadata.is_readable():
			msg = f'{readme} is not readable: {metadata.permissions}'
			raise PermissionError(msg)

		metadata.fields = cls._parse_readme(readme)
		# remove the content field from the fields dict
		metadata.content = metadata.fields.pop('content', '')

		return metadata

	@staticmethod
	def _find_readme(path: Path) -> Path:
		if path.is_dir():
			readmes = [f for f in path.glob('README*') if f.is_file()]
			if len(readmes) == 0:
				msg = 'No README file found.'
				raise ValueError(msg)
			elif len(readmes) > 1:
				readme = readmes[0]
			else:
				readme = readmes[0]
		else:
			readme = path

		if 'README' not in readme.stem.upper():
			msg = 'The file is not a README file.'
			raise ValueError(msg)

		return readme

	def _dirsize(self) -> ByteSize:
		size = get_directory_size(self.path)
		self.size = size
		self.size_measured_at = datetime.datetime.now().strftime('%Y-%m-%d-%H:%M')
		return size

	def write_dirsize(self) -> None:
		if not self.size:
			self._dirsize()
		dirsize_str = f'{self.size.B} - {self.size_measured_at}'
		self.fields['SIZE'] = dirsize_str
		self.write_to_file()

	def read_dirsize(self) -> ByteSize:
		dirsize_str = self.fields.get('SIZE', None)
		if dirsize_str:
			self.size = ByteSize(int(dirsize_str.split()[0]))
			self.size_measured_at = datetime.datetime.strptime(
				dirsize_str.split()[2], '%Y-%m-%d-%H:%M'
			)
			return self.size
		else:
			self.write_dirsize()
			return self.size

	def log_change(self, description: str) -> None:
		timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M')
		self.logs.append(f'{timestamp}: {description}')

	def add_field(self, field: str, value: str) -> None:
		"""Assuming the field is not already present, add it to the fields dict.

		If the field is already present, add the value to the existing field.

		Args:
		    field (str): name of the field, i.e OWNER, DATE, DESC, STATUS
		    value (str): value of the field
		"""
		field = field.upper()
		if field in self.fields:
			self.fields[field] += f' {value}'
		else:
			self.fields[field] = value

	def write_to_file(self, newpath: Path | None = None) -> None:
		newpath = newpath or self.readme
		newpath = newpath.resolve()
		# create the newpath file if it doesn't exist
		newpath.parent.mkdir(parents=True, exist_ok=True)
		newpath.touch(exist_ok=True)

		if not is_file_writable(newpath):
			msg = f'{newpath} is not writable: {self.permissions}'
			raise PermissionError(msg)

		# with newpath.open(mode='w') as file:
		file = newpath.open(mode='w')
		for fld, value in self.fields.items():
			line = f'#{fld}: {value}'
			if len(line) > 80:
				# If the line length exceeds 80,
				# split it into multiple lines without breaking words
				words = line.split()
				lines = []
				current_line = ''
				for word in words:
					if len(current_line) + len(word) + 1 <= 80:
						current_line += word + ' '
					else:
						lines.append(current_line.strip())
						current_line = word + ' '
				if current_line:
					lines.append(current_line.strip())
				file.write('\n'.join(lines))
			else:
				file.write(line)
			file.write('\n\n')
		if self.content:
			file.write(f'\n{self.content}')
		if self.logs:
			file.write('\n\n\n')
			for log in self.logs:
				file.write(f'\n{log}')
		file.close()

	def check_fields(self) -> None:
		missing = [fld for fld in MANDATORY_FIELDS if fld not in self.fields]
		if missing:
			msg = f'The following fields are missing: {missing} in {self.readme}'
			raise ValueError(msg)

	@classmethod
	def _parse_readme(
		cls: Type['DMPMetadata'],
		readme: Path,
		pattern: re.Pattern[str] = re.compile(r'^#([A-Z]+): (.+)$'),
	) -> dict:
		current_field = None
		current_value = []
		content_lines = []
		metadata = {}
		with readme.open(mode='r') as file:
			for line in file:
				if line.strip() == '' and current_field:
					# End current field on double newline
					metadata[current_field] = ' '.join(current_value).strip()
					current_field = None
					current_value = []
				else:
					match = pattern.match(line.strip())
					if match:
						if current_field:
							metadata[current_field] = ' '.join(current_value).strip()
						current_field, current_value = (
							match.groups()[0],
							[match.groups()[1]],
						)
					elif current_field:
						current_value.append(line.strip())
					else:
						content_lines.append(line.strip())

			if current_field:
				metadata[current_field] = ' '.join(current_value).strip()

		metadata['content'] = '\n'.join(content_lines).strip()
		return metadata

	def __getitem__(self, item: str) -> str:
		return self.fields.get(item, None)

	def __setitem__(self, key: str, value: str) -> None:
		self.fields[key] = value

	@staticmethod
	def evaluate_permissions(path: Path) -> str:
		permissions = path.stat().st_mode
		is_dir = 'd' if stat.S_ISDIR(permissions) else '-'
		perm_bits = [
			(permissions & stat.S_IRUSR, 'r'),
			(permissions & stat.S_IWUSR, 'w'),
			(permissions & stat.S_IXUSR, 'x'),
			(permissions & stat.S_IRGRP, 'r'),
			(permissions & stat.S_IWGRP, 'w'),
			(permissions & stat.S_IXGRP, 'x'),
			(permissions & stat.S_IROTH, 'r'),
			(permissions & stat.S_IWOTH, 'w'),
			(permissions & stat.S_IXOTH, 'x'),
		]
		formatted_permissions = is_dir + ''.join(
			bit[1] if bit[0] else '-' for bit in perm_bits
		)
		return formatted_permissions

	def get_permissions(self) -> str:
		if not self.permissions:
			return 'No permissions set.'
		return self.permissions

	def is_writeable(self) -> bool:
		return 'w' in self.permissions

	def is_readable(self) -> bool:
		return 'r' in self.permissions

	def __rich_repr__(self) -> rich.repr.Result:
		yield 'path', self.path
		yield 'fields', self.fields
		yield 'content', self.content
		yield 'permissions', self.permissions
		yield 'logs', self.logs
		yield 'size', self.size
		yield 'size_measured_at', self.size_measured_at



from pathlib import Path

import rich_click as click
from rich import print

from damply.metadata import MANDATORY_FIELDS, DMPMetadata


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.argument(
	'field',
	type=str,
)
@click.argument(
	'value',
	type=str,
)
def add_field(path: Path, field: str, value: str) -> None:
	"""Add a field to the README file."""

	if field.upper() in MANDATORY_FIELDS:
		print(
			f'Error: The field {field} is a mandatory field. Use the config command to modify it.'
		)
		return
	try:
		metadata = DMPMetadata.from_path(path)
		metadata.check_fields()
	except ValueError as e:
		print(f'Error: No log entry added.\n{e}')
		return

	metadata.add_field(field, value)
	metadata.log_change(f'Added {field} to README file.')
	metadata.write_to_file()



import platform
from pathlib import Path
from typing import NoReturn


def _raise_windows_not_supported() -> NoReturn:
	msg = 'Platform not supported for retrieving user info using pwd module.'
	raise NotImplementedError(msg)


def get_file_owner_full_name(file_path: Path) -> str:
	try:
		if platform.system() == 'Windows':
			_raise_windows_not_supported()

		from pwd import getpwuid

		# Get the file's status
		file_stat = file_path.stat()
		# Get the user ID of the file owner
		uid: int = file_stat.st_uid

		# Get the user information based on the user ID
		getpwuid(uid)

	except ImportError:
		return file_path.owner()
	except NotImplementedError:
		return 'Retrieving user info is not supported on Windows.'
	except Exception as e:
		return str(e)
	else:
		return getpwuid(uid).pw_gecos



from pathlib import Path

import click
from rich import print

from damply.audit import DirectoryAudit


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.option(
	'--json',
	is_flag=True,
	help='Output the audit results in JSON format.',
	default=False,
)
def audit(path: Path, json: bool) -> None:
	"""Audit the metadata of a valid DMP Directory."""

	try:
		audit = DirectoryAudit.from_path(path)
		if json:
			click.echo(audit.to_json())
		else:
			print(audit)
	except ValueError as e:
		print(e)
		return



import grp
from dataclasses import asdict, dataclass


# custom exception for group not found
class GroupNotFoundError(Exception):
	def __init__(self, group_name: str) -> None:
		super().__init__(f"Group '{group_name}' not found.")
		self.group_name = group_name


@dataclass
class GroupInfo:
	name: str
	gid: int
	members: list[str]

	@classmethod
	def from_group_name(cls, group_name: str) -> 'GroupInfo':
		try:
			group_data = grp.getgrnam(group_name)
		except KeyError as e:
			raise GroupNotFoundError(group_name) from e
		else:
			return cls(
				name=group_data.gr_name,
				gid=group_data.gr_gid,
				members=group_data.gr_mem,
			)

	@classmethod
	def from_group_id(cls, group: int) -> 'GroupInfo':
		try:
			group_data = grp.getgrgid(group)
		except KeyError as e:
			msg = f'GID {group}'
			raise GroupNotFoundError(msg) from e
		else:
			return cls(
				name=group_data.gr_name,
				gid=group_data.gr_gid,
				members=group_data.gr_mem,
			)

	def to_dict(self) -> dict:
		return asdict(self)

	def to_json(self, indent: int = 4) -> str:
		import json

		return json.dumps(self.to_dict(), indent=indent)


def get_all_groups() -> list[GroupInfo]:
	"""Returns a list of all groups on the system."""
	all_groups = []
	for group_data in grp.getgrall():
		all_groups.append(GroupInfo.from_group_id(group_data.gr_gid))
	return all_groups


if __name__ == '__main__':
	pass



from dataclasses import dataclass
from pathlib import Path
from typing import List

"""
This module contains functions to find symlinks in a directory.
For the most part, projects implementing the standard should use symbolic links when possible.

i.e

|_ PROJECT_NAME
|_ rawdata -> /path_to_rawdata_dir/PROJECT_NAME/
|_ procdata -> /path_to_procdata_dir/PROJECT_NAME/
|_ references -> /path_to_references_dir/GENCODE/...

OR

|_ PROJECT_NAME
|_ rawdata
  |_ rnaseq -> /path_to_rawdata_dir/PROJECT_NAME/rnaseq/
  |_ methylation -> /path_to_rawdata_dir/PROJECT_NAME/methylation/
|_ procdata
  |_ rnaseq -> /path_to_procdata_dir/PROJECT_NAME/rnaseq/
  |_ methylation -> /path_to_procdata_dir/PROJECT_NAME/methylation/

We should be able to determine quickly if a directory has any symlinks up to 2 levels deep.
"""


@dataclass
class SymlinkInfo:
	path: Path
	target: Path


def find_symlinks(directory: Path, max_depth: int = 3) -> List[SymlinkInfo]:
	"""
	Find symlinks in a directory up to a specified depth.

	Args:
	    directory (Path): The directory to search in.
	    max_depth (int): The maximum depth to search for symlinks.

	Returns:
	    List[SymlinkInfo]: A list of SymlinkInfo objects containing the symlink and its target.
	"""
	symlinks = []

	def search_dir(current_path: Path, current_depth: int) -> None:
		if current_depth > max_depth:
			return

		if current_path.is_symlink():
			symlinks.append(
				SymlinkInfo(path=current_path, target=current_path.resolve())
			)

		if current_path.is_dir():
			for entry in current_path.iterdir():
				search_dir(entry, current_depth + 1)

	search_dir(directory, 0)
	return symlinks


# Example usage
if __name__ == '__main__':
	project_dir = Path('/cluster/projects/bhklab/projects/BTCIS')
	symlinks = find_symlinks(project_dir)
	for _symlink in symlinks:
		pass



import pytest
from unittest.mock import patch, MagicMock
from pathlib import Path
import platform

from damply import whose


@pytest.mark.skipif(platform.system() != "Linux", reason="Test specific to Linux")
def test_get_file_owner_full_name_linux():
    file_path = Path('/dummy/path')

    with patch('os.stat') as mock_stat, patch('pwd.getpwuid') as mock_getpwuid:
        mock_stat.return_value.st_uid = 1000
        mock_user_info = MagicMock()
        mock_user_info.pw_gecos = 'John Doe'
        mock_getpwuid.return_value = mock_user_info

        assert whose.get_file_owner_full_name(file_path) == 'John Doe'


@pytest.mark.skipif(platform.system() != "Darwin", reason="Test specific to macOS")
def test_get_file_owner_full_name_macos():
    file_path = Path('/dummy/path')

    with patch('os.stat') as mock_stat, patch('pwd.getpwuid') as mock_getpwuid:
        mock_stat.return_value.st_uid = 501  # Common uid for first user in macOS
        mock_user_info = MagicMock()
        mock_user_info.pw_gecos = 'Mac User'
        mock_getpwuid.return_value = mock_user_info

        assert whose.get_file_owner_full_name(file_path) == 'Mac User'


@pytest.mark.skipif(platform.system() != "Windows", reason="Test specific to Windows")
def test_get_file_owner_full_name_windows():
    file_path = Path('C:\\dummy\\path')

    with patch('platform.system', return_value='Windows'):
        # On Windows, the function should return the message about Windows not being supported
        result = whose.get_file_owner_full_name(file_path)
        assert result == 'Retrieving user info is not supported on Windows.'


#!/usr/bin/env python3
# filepath: /Users/bhklab/dev/GitHub/damply/tests/test_dmpdirs.py
"""Tests for the dmpdirs module."""

import os
import sys
import platform
import shutil
import tempfile
from pathlib import Path
from unittest import mock

import pytest

from damply.dmpdirs import DamplyDirs, dirs, get_project_root
from damply.dmpdirs.exceptions import (
    DirectoryNameNotFoundError,
    EnvironmentVariableNotSetError,
)


@pytest.fixture
def temp_project_dir():
    """Create a temporary directory to use as a project root."""
    # Create a temp directory
    temp_dir = tempfile.mkdtemp(prefix="damply_test_")
    try:
        yield Path(temp_dir)
    finally:
        # Clean up - remove the directory after the test completes
        shutil.rmtree(temp_dir, ignore_errors=True)


@pytest.fixture
def reset_dirs_singleton():
    """Reset the DamplyDirs singleton after each test."""
    # Save the original state
    original_instance = DamplyDirs._instance
    original_initialized = original_instance._initialized if original_instance else None
    original_strict = original_instance._strict if original_instance else None
    
    # Reset for test
    DamplyDirs._instance = None
    
    yield
    
    # Restore original state after test
    DamplyDirs._instance = original_instance
    if original_instance:
        original_instance._initialized = original_initialized
        original_instance._strict = original_strict


@pytest.fixture
def setup_env_vars(temp_project_dir):
    """Set up environment variables for testing."""
    old_env = os.environ.copy()
    
    # Set up environment variables
    os.environ["DMP_PROJECT_ROOT"] = str(temp_project_dir)
    os.environ["RAWDATA"] = str(temp_project_dir / "custom_rawdata")
    
    yield
    
    # Restore environment variables
    os.environ.clear()
    os.environ.update(old_env)


class TestGetProjectRoot:
    """Tests for the get_project_root function."""
    
    def test_get_project_root_env_var(self, temp_project_dir):
        """Test get_project_root uses the DMP_PROJECT_ROOT environment variable."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            result = get_project_root()
            # Use resolve() on both sides to handle macOS /private prefix
            assert result.resolve() == temp_project_dir.resolve()
    
    def test_get_project_root_pixi_env_var(self, temp_project_dir):
        """Test get_project_root uses the PIXI_PROJECT_ROOT environment variable."""
        with mock.patch.dict(os.environ, {"PIXI_PROJECT_ROOT": str(temp_project_dir)}):
            assert get_project_root().resolve() == temp_project_dir.resolve()
            
    def test_get_project_root_cwd_fallback(self):
        """Test get_project_root falls back to current working directory."""
        with mock.patch.dict(os.environ, {}, clear=True):
            assert get_project_root() == Path.cwd().resolve()


class TestDamplyDirsSingleton:
    """Tests for the DamplyDirs singleton pattern."""
    
    def test_singleton_pattern(self, reset_dirs_singleton):
        """Test that DamplyDirs follows the singleton pattern."""
        dirs1 = DamplyDirs()
        dirs2 = DamplyDirs()
        assert dirs1 is dirs2
        
    def test_singleton_dir_cache(self, reset_dirs_singleton, temp_project_dir):
        """Test that the dir cache is properly maintained."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            dirs1 = DamplyDirs()
            dirs1.set_strict_mode(False)
            # Access a property to cache it
            _ = dirs1.RAWDATA
            
            # Create a second instance, which should be the same object
            dirs2 = DamplyDirs()
            
            # The cached path should be available in the second instance
            assert "RAWDATA" in dirs2._dir_cache
            assert dirs2._dir_cache["RAWDATA"] == dirs1._dir_cache["RAWDATA"]


class TestDamplyDirsDirectoryCreation:
    """Tests for directory creation behavior."""
    
    def test_strict_mode_directory_access(self, reset_dirs_singleton, temp_project_dir):
        """Test that strict mode raises an error when directories don't exist."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            # Initialize with strict mode enabled
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(True)
            
            # In strict mode, it should raise EnvironmentVariableNotSetError when env var is not set
            with pytest.raises(EnvironmentVariableNotSetError):
                _ = test_dirs.RAWDATA
    
    def test_non_strict_mode_directory_creation(self, reset_dirs_singleton, temp_project_dir):
        """Test that non-strict mode creates missing directories."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            # Initialize with strict mode disabled
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            
            # This should create the directory
            rawdata_dir = test_dirs.RAWDATA
            
            # Check that the directory was created
            assert rawdata_dir.exists()
            assert rawdata_dir.resolve() == (temp_project_dir / "data" / "rawdata").resolve()
    
    def test_set_strict_mode(self, reset_dirs_singleton, temp_project_dir):
        """Test the set_strict_mode method."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            assert test_dirs._strict is False
            
            test_dirs.set_strict_mode(True)
            assert test_dirs._strict is True
            
            # Should now raise an error for non-existent directories
            # First, let's ensure a directory doesn't exist
            results_path = temp_project_dir / "data" / "results"
            if results_path.exists():
                shutil.rmtree(results_path)
            
            # In strict mode, it should raise EnvironmentVariableNotSetError when env var is not set
            with pytest.raises(EnvironmentVariableNotSetError):
                _ = test_dirs.RESULTS


class TestDamplyDirsPathResolution:
    """Tests for path resolution behavior."""
    
    def test_path_from_env_var(self, reset_dirs_singleton, setup_env_vars, temp_project_dir):
        """Test that paths are resolved from environment variables when available."""
        test_dirs = DamplyDirs()
        rawdata_path = test_dirs.RAWDATA
        
        # Should use the env var path
        assert rawdata_path.resolve() == (temp_project_dir / "custom_rawdata").resolve()
        
    def test_default_paths(self, reset_dirs_singleton, temp_project_dir):
        """Test default path resolution when environment variables are not set."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            
            # Test each directory type
            assert test_dirs.RAWDATA.resolve() == (temp_project_dir / "data" / "rawdata").resolve()
            assert test_dirs.PROCDATA.resolve() == (temp_project_dir / "data" / "procdata").resolve()
            assert test_dirs.RESULTS.resolve() == (temp_project_dir / "data" / "results").resolve()
            assert test_dirs.CONFIG.resolve() == (temp_project_dir / "config").resolve()
            assert test_dirs.METADATA.resolve() == (temp_project_dir / "metadata").resolve()
            assert test_dirs.LOGS.resolve() == (temp_project_dir / "logs").resolve()
            assert test_dirs.SCRIPTS.resolve() == (temp_project_dir / "workflow" / "scripts").resolve()
            assert test_dirs.NOTEBOOKS.resolve() == (temp_project_dir / "workflow" / "notebooks").resolve()
            assert test_dirs.PROJECT_ROOT.resolve() == temp_project_dir.resolve()
    
    def test_strict_env_var_requirement(self, reset_dirs_singleton, temp_project_dir):
        """Test that strict mode requires env vars to be set."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(True)
            
            # Should raise an error because RAWDATA env var is not set and we're in strict mode
            with pytest.raises(EnvironmentVariableNotSetError):
                _ = test_dirs.RAWDATA


class TestDamplyDirsAccess:
    """Tests for different ways to access directories."""
    
    def test_attribute_access(self, reset_dirs_singleton, temp_project_dir):
        """Test attribute-style access to directories."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            
            # Test attribute access
            assert test_dirs.RAWDATA.resolve() == (temp_project_dir / "data" / "rawdata").resolve()
    
    def test_dictionary_access(self, reset_dirs_singleton, temp_project_dir):
        """Test dictionary-style access to directories."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            
            # Test dictionary access
            assert test_dirs["RAWDATA"].resolve() == (temp_project_dir / "data" / "rawdata").resolve()
    
    def test_invalid_directory_name(self, reset_dirs_singleton):
        """Test that accessing an invalid directory name raises an error."""
        test_dirs = DamplyDirs()
        
        # Test attribute access with invalid name
        with pytest.raises(AttributeError):
            _ = test_dirs.INVALID_DIR
        
        # Test dictionary access with invalid name
        with pytest.raises(KeyError):
            _ = test_dirs["INVALID_DIR"]


@pytest.mark.skipif(
    platform.system() != "Windows", 
    reason="This test is only relevant on Windows"
)
class TestWindowsSpecific:
    """Windows-specific tests."""
    
    def test_windows_paths(self, reset_dirs_singleton, temp_project_dir):
        """Test path handling on Windows."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            
            # On Windows, check that backslashes are handled correctly
            rawdata_dir = test_dirs.RAWDATA
            assert "\\" in str(rawdata_dir)
            assert rawdata_dir.exists()


@pytest.mark.skipif(
    platform.system() == "Windows", 
    reason="This test is only relevant on Unix-like systems"
)
class TestUnixSpecific:
    """Unix-specific tests."""
    
    def test_unix_symlinks(self, reset_dirs_singleton, temp_project_dir):
        """Test handling of symlinks on Unix-like systems."""
        # Create a real directory to link to
        real_dir = temp_project_dir / "real_rawdata"
        real_dir.mkdir(parents=True)
        
        # Create a symlink
        symlink_path = temp_project_dir / "data" / "rawdata"
        symlink_path.parent.mkdir(parents=True, exist_ok=True)
        
        # Use os.symlink directly to handle potential permissions issues
        try:
            os.symlink(str(real_dir), str(symlink_path))
        except OSError:
            pytest.skip("Unable to create symlink - might need elevated permissions")
            
        # Use non-strict mode and also set RAWDATA env var to make the test more robust
        with mock.patch.dict(os.environ, {
            "DMP_PROJECT_ROOT": str(temp_project_dir),
            "RAWDATA": str(symlink_path)
        }):
            test_dirs = DamplyDirs()
            
            # Should correctly resolve the symlink
            assert test_dirs.RAWDATA.resolve() == real_dir.resolve()


class TestInstanceRepr:
    """Tests for the string representation of directories."""
    
    def test_repr_format(self, reset_dirs_singleton, temp_project_dir):
        """Test the __repr__ method format."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            test_dirs = DamplyDirs()
            test_dirs.set_strict_mode(False)
            
            # Create a couple of directories to ensure they show up in repr
            _ = test_dirs.RAWDATA
            _ = test_dirs.CONFIG
            
            repr_str = repr(test_dirs)
            
            # Check that the repr contains expected elements
            assert "Project Root:" in repr_str
            assert "Strict Mode: OFF" in repr_str
            assert "RAWDATA" in repr_str
            assert "CONFIG" in repr_str


class TestImportedDirs:
    """Tests for the pre-imported dirs object."""
    
    def test_imported_dirs_instance(self):
        """Test that the imported dirs object is a DamplyDirs instance."""
        assert isinstance(dirs, DamplyDirs)
    
    def test_imported_dirs_attributes(self, temp_project_dir):
        """Test that the imported dirs object has the expected attributes."""
        with mock.patch.dict(os.environ, {"DMP_PROJECT_ROOT": str(temp_project_dir)}):
            # Reset dirs for this test
            dirs._dir_cache = {}
            dirs._project_root = get_project_root()
            
            # Test that we can access the attributes
            assert hasattr(dirs, "RAWDATA")
            assert hasattr(dirs, "set_strict_mode")
            
            # Test dir() includes the valid directories
            dir_list = dir(dirs)
            for valid_dir in dirs._VALID_DIRS:
                assert valid_dir in dir_list


if __name__ == "__main__":
    pytest.main(["-xvs", __file__])


import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Optional

# ruff: noqa


@dataclass
class DirectoryModificationMetadata:
	directory: Path
	last_modification_time: Optional[int] = None
	last_modification_time_measured_at: Optional[datetime] = None

	def __post_init__(self):
		# Ensure the directory exists and is a directory
		self.validate_directory()

		# Initialize modification time and the measurement timestamp
		self.update_last_modification_time(initial=True)

	def validate_directory(self) -> None:
		"""Validate that the directory exists and is indeed a directory."""
		if not self.directory.exists():
			msg = f"Directory '{self.directory}' does not exist."
			raise FileNotFoundError(msg)
		if not self.directory.is_dir():
			msg = f"Path '{self.directory}' is not a directory."
			raise NotADirectoryError(msg)

	def get_last_modification_time(self) -> int:
		"""Recursively get the last modification time of the directory and its contents."""
		# if empty, modification time is just time of directory's actual modification and nothing to do with files
		last_mod_time = max(
			(f.stat().st_mtime for f in self.directory.rglob('*') if f.is_file()),
			default=self.directory.stat().st_mtime,
		)
		return int(last_mod_time)

	def update_last_modification_time(self, initial: bool = False) -> None:
		"""Update the stored last modification time and measurement timestamp."""
		self.last_modification_time = self.get_last_modification_time()
		self.last_modification_time_measured_at = datetime.now()
		if not initial:
			pass

	def update(self) -> None:
		"""Update the metadata if the directory has been modified since the last check."""
		if self.get_last_modification_time() != self.last_modification_time:
			self.update_last_modification_time(initial=True)

		self.get_file_count()
		self.get_total_size()
		self.has_subdirectories()

	def get_file_count(self) -> int:
		"""Get the total number of files in the directory."""
		return sum(1 for _ in self.directory.rglob('*') if _.is_file())

	def get_total_size(self) -> int:
		"""Calculate the total size of the directory."""
		return sum(f.stat().st_size for f in self.directory.rglob('*') if f.is_file())

	def has_subdirectories(self) -> bool:
		"""Check if the directory contains any subdirectories."""
		return any(f.is_dir() for f in self.directory.iterdir())

	def display(self) -> None:
		"""Display the metadata and check for modifications."""
		from rich.console import Console
		from rich.panel import Panel
		from rich.table import Table

		self.update()

		console = Console()
		table = Table(title=f'Directory: {self.directory.absolute()}')
		table.add_column('Property', justify='left', style='cyan', no_wrap=True)
		table.add_column('Value', justify='left', style='magenta')

		table.add_row(
			'Last modification time',
			str(datetime.fromtimestamp(self.last_modification_time)),
		)
		table.add_row(
			'Last modification time measured at',
			str(self.last_modification_time_measured_at),
		)
		table.add_row('File count', str(self.get_file_count()))
		table.add_row('Total size', str(self.get_total_size()))
		table.add_row('Has subdirectories', str(self.has_subdirectories()))

		panel = Panel(table, title='Directory Metadata', border_style='red')
		console.print(panel)


if __name__ == '__main__':
	# directory_path = Path("/path/to/your/directory")
	import tempfile
	import shutil
	import time
	from loguru import logger

	logger.info('Starting Directory Modification Metadata Test')

	directory_path = Path(tempfile.mkdtemp())

	logger.info(f'Created temporary directory: {directory_path}')
	metadata = DirectoryModificationMetadata(directory=directory_path)
	metadata.display()

	logger.info('Sleeping for 5 seconds to allow for initial modification time capture')
	time.sleep(5)
	# Create a file in the directory
	logger.info('Creating a new file in the directory')
	with open(os.path.join(directory_path, 'new_file.txt'), 'w') as f:
		f.write('New file content')

	metadata.display()

	logger.info('Sleeping for 2 seconds to allow for file creation')
	time.sleep(2)
	# Delete the file
	logger.info('Deleting the file created in the directory')
	os.remove(os.path.join(directory_path, 'new_file.txt'))
	metadata.display()

	logger.info('Removing the temporary directory')
	shutil.rmtree(directory_path)

# def test_complex_multiple_subdirectories():
#     """
#     This test will do the following:
#     1. Create a directory structure with multiple subdirectories
#     2. Create a file in each subdirectory
#     3. Record the last modification time of each subdirectory
#     4. Add a file to each subdirectory
#     5. Record the last modification time of each subdirectory. Ensure that the last modification time has changed.
#     6. Update one of the files in each subdirectory
#     7. Record the last modification time of each subdirectory. Ensure that the last modification time has changed.
#     """
#     import os
#     import shutil
#     import time
#     import tempfile

#     base_dir = tempfile.mkdtemp()
#     subdirs = [f'subdir_{i}' for i in range(5)]
#     files = [f'file_{i}.txt' for i in range(5)]

#     if os.path.exists(base_dir):
#         shutil.rmtree(base_dir)
#     os.mkdir(base_dir)

#     initial_times = {}
#     updated_times_after_add = {}
#     final_times_after_update = {}

#     try:
#         # Step 1: Create subdirectories
#         for subdir in subdirs:
#             os.mkdir(os.path.join(base_dir, subdir))

#         # Step 2: Create a file in each subdirectory
#         for subdir, file_name in zip(subdirs, files):
#             with open(os.path.join(base_dir, subdir, file_name), 'w') as f:
#                 f.write('Initial content')

#         # Step 3: Record the last modification time of each subdirectory and ensure it changed
#         dmm_list = [DirectoryModificationMetadata(directory=Path(os.path.join(base_dir, subdir))) for subdir in subdirs]
#         for dmm in dmm_list:
#             dmm.display()
#             assert dmm.get_last_modification_time() > 0, f"Modification time is not greater than 0 for {dmm.directory.absolute()}"
#         time.sleep(5)
#         print("\n\n")
#         # Step 4: Add a file to each subdirectory
#         for subdir in subdirs:
#             new_file_path = os.path.join(base_dir, subdir, 'new_file.txt')
#             with open(new_file_path, 'w') as f:
#                 f.write('New file content')

#         # Step 5: Record the last modification time of each subdirectory and ensure it changed
#         for dmm in dmm_list:
#             dmm.display()
#             assert dmm.get_last_modification_time() > 0, f"Modification time is not greater than 0 for {dmm.directory.absolute()}"

#         time.sleep(5)
#         print("\n\n")

#         # Check again
#         for dmm in dmm_list:
#             dmm.display()


#     finally:
#         # Cleanup
#         if os.path.exists(base_dir):
#             shutil.rmtree(base_dir)



"""cli entry to find the user who owns a file or directory."""

from pathlib import Path

import click

from damply.admin import UserInfo


class WindowsNotSupportedError(NotImplementedError):
	"""Exception raised when trying to use pwd module on Windows."""

	pass


@click.command()
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.option(
	'--json',
	'-j',
	is_flag=True,
	help='Output in JSON format.',
)
def whose(path: Path, json: bool) -> None:
	"""Print the owner of the file or directory."""
	import platform

	if platform.system() == 'Windows':
		msg = "The 'whose' command is not supported on Windows. "
		raise WindowsNotSupportedError(msg)

	# get the files owner
	file_owner = path.stat().st_uid

	# get the user info
	user_info = UserInfo.from_uid(file_owner)
	if user_info is None:
		click.echo(f'No user found for UID {file_owner}.')
		return

	if json:
		info_dict = {
			'path': str(path.resolve()),
			'username': user_info.name,
			'uid': user_info.uid,
			'realname': user_info.realname,
		}
		import json

		click.echo(json.dumps(info_dict, indent=2))
	else:
		click.echo(f'Path: {path.resolve()}')
		click.echo(f'Username: {user_info.name}')
		click.echo(f'UID: {user_info.uid}')
		click.echo(f'Real Name: {user_info.realname}')



from datetime import datetime
from pathlib import Path
from typing import List

import pandas as pd
import plotly.graph_objects as go

from damply.utils import Directory, DirectoryList

"""This is a sandbox script to plot a sankey diagram of the dmp"""

MANDATORY_COLUMNS = ['abspath', 'size_GB']

"""
The input file should contain the mandatory columns and optionally other columns.
The mandatory columns are:
- abspath: the absolute path of the directory
- size_GB: the size of the directory in GB

Column Headings:
owner	full_name	abspath	relative_path	directory	parent_dir	permissions	creation_date
last_modified	has_README	OWNER	DATE	DESC	size_MB	size_GB

Goal:
"""


def permutate_path(path: Path) -> List[Path]:
	"""
	Given a path, return all possible paths from the root to the path
	"""
	assert path.is_absolute(), 'The path must be absolute'

	nodes = []
	while path != Path('/'):
		nodes.append(path)
		path = path.parent
	return nodes


def generate_node_list(dirlist: DirectoryList) -> List[Path]:
	common_root = dirlist.common_root
	nodes = set(permutate_path(common_root))

	for directory in dirlist.directories:
		nodes.update(permutate_path(directory.directory))
	# sort the nodes by # of "/" in the path and then alphabetically
	nodes = sorted(nodes, key=lambda x: (len(x.parts), x))
	return list(nodes)


def damplyplot(
	file_path: Path,
	threshold_gb: int = 100,
	fig_width: int = 3340,
	fig_height: int = 1440,
	depth_from_common_root: int = 3,
) -> Path:
	"""

	The goal is to create a sankey diagram of the directories where the source of
	each flow is the parent directory and the target is the child directory with
	the width of the flow being the size of the directory
	"""

	# Read the file
	audit_df = pd.read_csv(file_path, sep='\t')

	if not all(col in audit_df.columns for col in MANDATORY_COLUMNS):
		msg = f'The file must contain the following columns: {", ".join(MANDATORY_COLUMNS)}'
		raise ValueError(msg)

	# Filter the dataframe
	audit_df = audit_df[audit_df['size_GB'] > threshold_gb]

	dirlist: DirectoryList = DirectoryList(
		directories=[
			Directory(directory=Path(row['abspath']), size_GB=row['size_GB'])
			for index, row in audit_df.iterrows()
		]
	)
	nodes = generate_node_list(dirlist)

	links = []
	for _dir in dirlist:
		target = nodes.index(_dir.directory)
		source = nodes.index(_dir.parent)
		links.append({'source': source, 'target': target, 'value': _dir.size_GB})

	label_with_sizes = []
	for node in nodes:
		label = node.name
		size = dirlist.dir_size_dict().get(node, 0)

		if size == 0:
			children = [child for child in dirlist.directories if child.parent == node]
			size = sum([child.size_GB for child in children])

			# add to dirlist
			dirlist.directories.append(Directory(directory=node, size_GB=size))

		label_with_sizes.append(f'{label} ({size} GB)')

	nodes_whose_parent_is_common_root = [
		node for node in nodes if node.parent == dirlist.common_root
	]

	# add a link from the common root to the nodes whose parent is the common root
	common_root_size = sum(
		[
			_dir.size_GB
			for _dir in dirlist.directories
			if _dir.parent == dirlist.common_root
		]
	)
	for node in nodes_whose_parent_is_common_root:
		target = nodes.index(node)
		source = nodes.index(dirlist.common_root)
		size = dirlist.dir_size_dict().get(node, 0)
		links.append({'source': source, 'target': target, 'value': size})

	# get the index of the common root so that we can update the label_with_sizes index
	common_root_index = nodes.index(dirlist.common_root)
	label_with_sizes[common_root_index] = (
		f'{dirlist.common_root} ({common_root_size} GB)'
	)

	fig_layout = {'width': fig_width, 'height': fig_height}

	fig = go.Figure(
		data=[
			go.Sankey(
				node={
					'pad': 30,
					'thickness': 10,
					'line': {'color': 'black', 'width': 0.5},
					'label': label_with_sizes,
					'color': 'blue',
				},
				link={
					'source': [link['source'] for link in links],
					'target': [link['target'] for link in links],
					'value': [link['value'] for link in links],
				},
				textfont={'color': 'black', 'size': 20},
			)
		],
		layout=fig_layout,
	)

	# save the figure using todays date as damplyplot_{MM-DD-YYYY}.png
	date_str = datetime.now().strftime('%m-%d-%Y')

	output_path = Path(f'damplyplot_{date_str}.png')
	fig.write_image(output_path)

	return output_path



from pathlib import Path

import rich_click as click
from rich import print

from damply import __version__
from damply.cli.add_field import add_field
from damply.cli.audit import audit
from damply.cli.click_config import help_config
from damply.cli.init import init
from damply.cli.plot import plot
from damply.metadata import MANDATORY_FIELDS, DMPMetadata
from damply.utils import whose as whose_util


@click.group(
	name='damply',
	context_settings={'help_option_names': ['-h', '--help']},
)
@click.version_option(__version__, prog_name='damply')
def cli() -> None:
	"""
	A tool to interact with systems implementing the
	Data Management Plan (DMP) standard.

	This tool is meant to allow sys-admins to easily query and audit the metadata of their
	projects.
	"""
	pass


@cli.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'directory',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=False,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.rich_config(help_config=help_config)
def view(directory: Path) -> None:
	"""View the DMP Metadata of a valid DMP Directory."""
	try:
		metadata = DMPMetadata.from_path(directory)
		metadata.check_fields()
	except ValueError as e:
		print(e)
		return

	from rich.console import Console
	from rich.markdown import Markdown
	from rich.table import Table

	console = Console()

	table = Table.grid(padding=1, pad_edge=True, expand=True)
	table.title = f'[bold]Metadata for {metadata.path.absolute()}[/bold]'
	table.add_column('Field', justify='right', style='cyan')
	table.add_column('Value', style='yellow')

	for field, value in metadata.fields.items():
		table.add_row(field, value)

	console.print(table)
	console.print(Markdown(metadata.content))
	console.print(Markdown('\n'.join(metadata.logs)))


@cli.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.rich_config(help_config=help_config)
def whose(path: Path) -> None:
	"""Print the owner of the file or directory."""
	result = whose_util.get_file_owner_full_name(path)

	print(
		f'The owner of [bold magenta]{path}[/bold magenta] is [bold cyan]{result}[/bold cyan]'
	)


@cli.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option(
	'--path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.argument(
	'description',
	type=str,
)
@click.rich_config(help_config=help_config)
def log(description: str, path: Path) -> None:
	"""Add a log entry to the metadata."""
	try:
		metadata = DMPMetadata.from_path(path)
		metadata.check_fields()
	except ValueError as e:
		print(f'Error: No log entry added.\n{e}')
		return

	metadata.log_change(description)
	metadata.write_to_file()


@cli.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option(
	'--dry_run',
	is_flag=True,
	default=False,
)
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
@click.rich_config(help_config=help_config)
def config(path: Path, dry_run: bool) -> None:
	"""Modify or add the fields in the README file."""

	try:
		metadata = DMPMetadata.from_path(path)
		metadata.check_fields()
	except ValueError as e:
		print(f'{e}')

	from rich.console import Console

	if dry_run:
		config_path = Path(metadata.readme + '.dmp')
		print(
			f'Dry run mode is on. No changes will be made and'
			f'changes will be written to [bold cyan]{config_path.resolve()}[/bold cyan]'
		)
	else:
		config_path = metadata.readme

	# here, we show a prompt for each field in metadata.fields
	# and show the current value if it exists
	console = Console()
	for field in MANDATORY_FIELDS:
		value = metadata.fields.get(field, '[red]NOT SET[/red]')
		console.print(f'[bold]{field}[/bold]: [cyan]{value}[/cyan]')
		new_value = console.input(f'Enter a new value for {field}: ')

		if not new_value and not metadata.fields.get(field):
			print(f'[red]Field {field} MUST be set.[/red]')
			return
		metadata[field] = new_value

	metadata.log_change('Updated fields in README file.')

	if not dry_run:
		metadata.write_to_file(config_path)
		console.print(metadata)


@cli.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		path_type=Path,
		file_okay=True,
		dir_okay=True,
		readable=True,
	),
	default=Path().cwd(),
)
def size(path: Path) -> None:
	"""Print the size of the directory."""
	try:
		metadata = DMPMetadata.from_path(path)
		metadata.check_fields()
	except ValueError as e:
		print(f'{e}')
		return

	size_dir = metadata.read_dirsize()
	from rich.console import Console

	console = Console()

	console.print(
		f'The size of [bold magenta]{path}[/bold magenta] is [bold cyan]{size_dir}[/bold cyan]'
	)
	print(metadata)


cli.add_command(audit)
cli.add_command(plot)
cli.add_command(add_field)
cli.add_command(init)

if __name__ == '__main__':
	cli()



import pytest
from pathlib import Path
from damply.metadata import DMPMetadata
import datetime  # Add this import


def test_from_path_valid_readme():
    readme_path = Path("tests/examples/simple/README_simple.md").resolve()
    metadata = DMPMetadata.from_path(readme_path)
    assert metadata.fields["OWNER"] == "Jermiah Joseph"
    assert metadata.fields["DATE"] == "2024-05-30"
    assert metadata.fields["DESC"] == "A simple readme."

def test_from_path_invalid_readme():
    readme_path = Path("tests/examples/invalid_.md").resolve()
    with pytest.raises(ValueError):
        DMPMetadata.from_path(readme_path)


def test_log_change():
    metadata = DMPMetadata()
    metadata.log_change("Added a log entry.")
    metadata.log_change("Added another log entry.")
    # assert that the format is of:
    # timestamp1: Added a log entry.
    # timestamp2: Added another log entry.
    for log in metadata.logs:
        timestamp, message = log.split(": ")
        datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M")
        assert message.startswith("Added")


def test_write_to_file():
    metadata = DMPMetadata()
    metadata["FIELD1"] = "value1"
    metadata["FIELD2"] = "value2"
    metadata.content = "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
    metadata.log_change("Added a log entry.")
    metadata.log_change("Added another log entry.")

    tmpdirname = "tests/examples/"

    newpath = Path(tmpdirname) / "test.dmp"

    # create the file with all permissions
    newpath.touch()
    
    metadata.write_to_file(newpath)

    with newpath.open(mode="r") as file:
        content = file.read()

    for i, line in enumerate(content.split("\n")):
        if i == 0:
            assert line == "#FIELD1: value1"
        if i == 2:
            assert line == "#FIELD2: value2"
        if i == 5:
            assert line == "Lorem ipsum dolor sit amet, consectetur adipiscing elit."